"""
Damped Newton step
"""


#[

import numpy as _np
import scipy as _sp
from numbers import Real

from . import iterative as _iterative

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from typing import Any
    from . import iterative as _iterative

#]


_DAMPING_FACTOR = 0.9
_MIN_STEP = 0.05
_MIN_IMPROVEMENT = 0


def eval_step(
    guess: _iterative.GuessType,
    func: _iterative.ArrayType,
    jacob: _iterative.SparseArrayType,
    norm: float,
    eval_func: _iterative.FuncEvalType,
    eval_norm: _iterative.NormEvalType,
) -> tuple[_iterative.GuessType, _iterative.ArrayType, Real, ]:
    """
    """

    direction = - _sp.sparse.linalg.spsolve(jacob, func, )

    new_guess, new_func, new_step, = _damp_search(
        direction=direction,
        prev_guess=guess,
        prev_norm=norm,
        eval_func=eval_func,
        eval_norm=eval_norm,
    )

    return new_guess, new_func, new_step


def _damp_search(
    direction: _iterative.ArrayType,
    prev_guess: _iterative.GuessType,
    prev_norm: _iterative.ArrayType,
    eval_func: _iterative.FuncEvalType,
    eval_norm: _iterative.NormEvalType,
    ) -> tuple[_iterative.GuessType, _iterative.ArrayType, Real, ]:
    """
    """
    min_norm = (1 - _MIN_IMPROVEMENT) * prev_norm
    new_step = None

    while True:
        new_step = 1 if new_step is None else (_DAMPING_FACTOR * new_step)
        if new_step < _MIN_STEP:
            raise _iterative.StepFailure
        new_guess = prev_guess + new_step * direction
        new_func = eval_func(new_guess, )
        new_norm = eval_norm(new_func, )
        if new_norm < min_norm:
            break

    return new_guess, new_func, new_step,


# def _backtracking_line_search(
#     prev_guess: GuessType,
#     prev_func: ArrayType,
#     eval_func: FuncEvalType,
#     eval_norm: NormEvalType,
#     unit_step: ArrayType,
#     args: dict,
#     alpha: float = 1.0,
#     beta_init: float = 0.8,
#     beta_min: float = 0.1,
#     beta_max: float = 0.9,
#     c: float = 1e-4,
#     max_iter: int = 1000,
# ) -> tuple[float, bool]:
#     """
#     Paper: https://ar5iv.labs.arxiv.org/html/1904.06321
#     """
# 
#     # start off with an initial step size
#     step_size = alpha
#     beta = beta_init
#     iter = 0
# 
#     # define epsilon for the gradient approximation
#     epsilon = _np.sqrt(_np.finfo(float).eps)
# 
#     # calculate the initial norm of func
#     prev_func_norm = eval_norm(prev_func)
# 
#     while True:
#         # check current step
#         if iter == max_iter:
#             return step_size, False
# 
#         # calculate the new candidate and evualuate it
#         new_guess = prev_guess + step_size * unit_step
#         new_func = eval_func(new_guess, args["data"])
# 
#         # calculate the gradient
#         prev_grad = _sp.optimize.approx_fprime(
#             prev_guess,
#             eval_func,
#             epsilon,
#             args["data"],
#         )
# 
#         # calculate the decrease
#         decrease = c * step_size * _np.dot(prev_grad, unit_step)
# 
#         # calculate the norms
#         new_func_norm = eval_norm(new_func)
#         decrease_norm = eval_norm(decrease)
# 
#         # check the Armijo (sufficient decrease) condition
#         # if not met, update the step size
#         if new_func_norm <= prev_func_norm + decrease_norm:
#             return step_size, True
#         step_size *= beta
# 
#         # adjust beta dynamically based on function decrease
#         if new_func_norm > prev_func_norm:
#             # if function decrease is slow,
#             # reduce beta to make larger reductions in step size
#             beta = max(beta * 0.9, beta_min)
#         else:
#             # if function decrease is fast,
#             # increase beta to avoid making small step size changes
#             beta = min(beta * 1.1, beta_max)
# 
#         # prepare for the next iteration
#         prev_func_norm = new_func_norm
#         iter += 1



import irispie as ir
from neqs import *

# load the pickled data
pickle = ir.load("tests/evaluator.pkl")
evaluator = pickle["evaluator"]
data = pickle["data"]

test, status = quasi_newton(
    evaluator.eval_func,
    evaluator.eval_jacob,
    evaluator.get_init_guess(data),
    {
        "step_tolerance": 1e-4,
        "func_tolerance": 1e-4,
        "max_iterations": 1000,
        "norm_order": None,
    },
    {
        "data": data,
    },
)

print(f"Solver status: {status}")
print(f"Solver results: {test}")



"""
Iterative numerical solver boilerplate
"""


#[

import numpy as _np
import scipy as _sp
import enum as _en
import functools as _ft
from typing import Any, Callable

from .iter_printers import IterPrinter

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

#]


_DEFAULT_SOLVER_SETTINGS = {
    "step_tolerance": 1e-12,
    "func_tolerance": 1e-12,
    "max_iterations": 5_000,
    "norm_order": None,
    "eval_jacob_every": 1,
    "eval_jacob_last": None,
}


class StepFailure(Exception):
    pass


ArrayType = _np.ndarray
SparseArrayType = _sp.sparse.spmatrix
FuncEvalType = Callable[[ArrayType], ArrayType]
JacobEvalType = Callable[[ArrayType], SparseArrayType]
NormEvalType = Callable[[ArrayType], float]
SettingsType = dict[str, Any]
GuessType = dict[str, Any]


_DEFAULT_ITER_PRINTER_COLUMNS = (
    "state",
    "func_norm",
    "step_length",
    "jacob_status",
    "worst_diff_x",
    "worst_func",
)


class ExitStatus(_en.Enum, ):
    """
    """
    #[

    SUCCESS = 0, "Successfully completed",
    MAX_ITERATIONS = 1, "Maximum number of iterations reached",
    CANNOT_MAKE_FURTHER_PROGRESS = 2, "Cannot make further progress",
    ERROR_EVALUATING_STEP = 3, "Error when evaluating the next step",

    @property
    def is_success(self, ) -> bool:
        return self.value[0] == 0

    def int(self, ) -> int:
        return self.value[0]

    def __str__(self, ) -> str:
        return self.value[1]

    #]


def iterate(
    eval_step: Callable,
    *,
    eval_func: FuncEvalType,
    eval_jacob: JacobEvalType,
    init_guess: ArrayType,
    iter_printer: IterPrinter | None = None,
    args: tuple[Any, ...] = (),
    **kwargs,
) -> tuple[GuessType, ExitStatus]:
    """
    """

    solver_settings = _resolve_solver_settings(**kwargs, )

    args = args or ()

    if iter_printer is None:
        iter_printer = IterPrinter(
            **(iter_printer_settings or {}),
            columns=_DEFAULT_ITER_PRINTER_COLUMNS,
        )

    state = {
        "iter": 0,
        "func_eval": 0,
        "jacob_eval": 0,
        "jacob_status": False,
    }

    def _eval_func(guess: ArrayType, ) -> ArrayType:
        state["func_eval"]+= 1
        return eval_func(guess, *args, )

    def _eval_jacob(guess: ArrayType, jacob, ) -> SparseArrayType:
        eval_jacob_every_satistifed = state["iter"] % solver_settings["eval_jacob_every"] == 0 
        eval_jacob_last_satistifed = (
            state["iter"] <= solver_settings["eval_jacob_last"]
            if solver_settings["eval_jacob_last"] is not None
            else True
        )
        if eval_jacob_every_satistifed and eval_jacob_last_satistifed:
            jacob = None
        state["jacob_status"] = jacob is None
        if state["jacob_status"]:
            state["jacob_eval"]+= 1
            jacob = eval_jacob(guess, *args, )
        return jacob

    eval_norm = _ft.partial(
        _sp.linalg.norm,
        ord=solver_settings["norm_order"],
    )

    curr_guess = init_guess
    prev_guess = curr_guess
    curr_func = _eval_func(curr_guess, )
    curr_step = None
    curr_jacob = None

    while True:

        curr_norm = eval_norm(curr_func, )
        iter_printer.next(
            guess=curr_guess,
            func=curr_func,
            jacob_status=state["jacob_status"],
            step_length=curr_step,
        )

        convergence_status = _check_convergence(
            curr_guess=curr_guess,
            prev_guess=prev_guess,
            curr_norm=curr_norm,
            eval_norm=eval_norm,
            func_tolerance=solver_settings["func_tolerance"],
            step_tolerance=solver_settings["step_tolerance"],
        )

        if convergence_status:
            exit_status = ExitStatus.SUCCESS
            break

        if state["iter"] >= solver_settings["max_iterations"]:
            exit_status = ExitStatus.MAX_ITERATIONS
            break

        curr_jacob = _eval_jacob(curr_guess, curr_jacob, )
        prev_guess = curr_guess.copy()

        try:
            curr_guess, curr_func, curr_step, = eval_step(
                guess=curr_guess,
                func=curr_func,
                jacob=curr_jacob,
                norm=curr_norm,
                eval_func=_eval_func,
                eval_norm=eval_norm,
            )
        except StepFailure as exception:
            exit_status = ExitStatus.CANNOT_MAKE_FURTHER_PROGRESS
            break
        except StepFailure as exception:
            exit_status = ExitStatus.ERROR_EVALUATING_STEP
            break

        state["iter"] += 1

    iter_printer.conclude()
    return curr_guess, exit_status,


def _resolve_solver_settings(**kwargs, ) -> SettingsType:
    """
    """
    solver_settings = dict(**_DEFAULT_SOLVER_SETTINGS, )
    for n in solver_settings.keys():
        custom_value = kwargs.get(n, None, )
        if custom_value is not None:
            solver_settings[n] = custom_value
    return solver_settings


def _check_convergence(
    curr_guess,
    prev_guess,
    curr_norm,
    step_tolerance,
    func_tolerance,
    eval_norm,
) -> bool:
    """
    """
    step_norm = eval_norm(curr_guess - prev_guess)
    func_tolerance_satisfied = curr_norm < func_tolerance
    step_tolerance_satisfied = step_norm < step_tolerance
    return func_tolerance_satisfied and step_tolerance_satisfied




"""
"""


#[

import importlib.metadata as _md
import functools as _ft

from .iterative import iterate, ExitStatus
from . import damped_newton
from .iter_printers import IterPrinter

#]


damped_newton = _ft.partial(iterate, damped_newton.eval_step, )


__version__ = _md.version(__name__)
__doc__ = _md.metadata(__name__).json["description"]

